library state;

import 'package:built_collection/built_collection.dart';
import 'package:built_redux/built_redux.dart';
import 'package:built_value/built_value.dart';
import '../models/enums.dart';
import '../models/point.dart';
import '../models/tetromino.dart';

part 'state.g.dart';

abstract class GameStateActions extends ReduxActions {
  GameStateActions._();
  factory GameStateActions() => new _$GameStateActions();

  ActionDispatcher<Null> pause;
  ActionDispatcher<Null> resume;
  ActionDispatcher<Null> tick;
  ActionDispatcher<MoveDirection> movePiece;
  ActionDispatcher<RotateDirection> rotatePiece;
  ActionDispatcher<Null> forceDown;
}

abstract class GameState implements Built<GameState, GameStateBuilder> {
  GameState._();
  factory GameState() => new _$GameState._(
        speed: 0,
        isPaused: false,
        queue: new BuiltList<Tetromino>.build((b) => b
          ..add(new Tetromino.Random())
          ..add(new Tetromino.Random())
          ..add(new Tetromino.Random())
          ..add(new Tetromino.Random())),
        board: new BuiltList<BuiltList<Tetromino>>(
            new List<BuiltList<Tetromino>>.generate(
                10,
                (d) => new BuiltList<Tetromino>(
                    new List<Tetromino>.generate(40, (_) => null)))),
        gameOver: false,
      );

  int get speed;

  bool get isPaused;

  BuiltList<Tetromino> get queue;

  BuiltList<BuiltList<Tetromino>> get board;

  bool get gameOver;

  @memoized
  Tetromino get currentTetromino => queue.first;
}

Reducer<GameState, GameStateBuilder, dynamic> createGameStateReducer() =>
    (new ReducerBuilder<GameState, GameStateBuilder>()
          ..add(GameStateActionsNames.pause, _pause)
          ..add(GameStateActionsNames.resume, _resume)
          ..add(GameStateActionsNames.tick, _tick)
          ..add(GameStateActionsNames.movePiece, _movePiece)
          ..add(GameStateActionsNames.rotatePiece, _rotatePiece)
          ..add(GameStateActionsNames.forceDown, _forceDown))
        .build();

void _pause(GameState state, Action<Null> action, GameStateBuilder builder) {
  builder.isPaused = true;
}

void _resume(GameState state, Action<Null> action, GameStateBuilder builder) {
  builder.isPaused = false;
}

void _tick(GameState state, Action<Null> action, GameStateBuilder builder) {
  _moveDown(builder);
  if (!_canMoveDown(state)) _settlePiece(state, builder);
}

void _movePiece(
    GameState state, Action<MoveDirection> action, GameStateBuilder builder) {
  switch (action.payload) {
    case MoveDirection.Down:
      _moveDown(builder);
      if (!_canMoveDown(builder.build())) {
        _settlePiece(state, builder);
      }
      break;
    case MoveDirection.Left:
      if (_canMoveLeft(state)) {
        _moveLeft(builder);
        if (!_canMoveDown(builder.build())) {
          _settlePiece(state, builder);
        }
      }
      break;
    case MoveDirection.Right:
      if (_canMoveRight(state)) {
        _moveRight(builder);
        if (!_canMoveDown(builder.build())) {
          _settlePiece(state, builder);
        }
      }
      break;
  }
}

void _rotatePiece(GameState state, Action<RotateDirection> action,
    GameStateBuilder builder) {}

void _forceDown(
    GameState state, Action<Null> action, GameStateBuilder builder) {
  while (_canMoveDown(builder.build())) _moveDown(builder);
  _settlePiece(state, builder);
}

bool _canMoveDown(GameState state) => state.currentTetromino.points
    .map((point) => state.board[point.x][point.y + 1])
    .every((tetromino) =>
        tetromino == null || tetromino == state.currentTetromino);

bool _canMoveRight(GameState state) => state.currentTetromino.points
    .map((point) => state.board[point.x + 1][point.y])
    .every((tetromino) =>
        tetromino == null || tetromino == state.currentTetromino);

bool _canMoveLeft(GameState state) => state.currentTetromino.points
    .map((point) => state.board[point.x - 1][point.y])
    .every((tetromino) =>
        tetromino == null || tetromino == state.currentTetromino);

void _moveDown(GameStateBuilder builder) {
  _move(builder, (b) => b.y + 1);
}

void _moveRight(GameStateBuilder builder) {
  _move(builder, (b) => b.x + 1);
}

void _moveLeft(GameStateBuilder builder) {
  _move(builder, (b) => b.x - 1);
}

void _move(GameStateBuilder builder, void pointBuilder(PointBuilder builder)) {
  builder.queue[0].points
    ..forEach((point) => builder.board[point.x] =
        builder.board[point.x].rebuild((l) => l[point.y] = null))
    ..map((point) => point.rebuild(pointBuilder))
    ..forEach((point) => builder.board[point.x] =
        builder.board[point.x].rebuild((l) => l[point.y] = builder.queue[0]));
}

void _settlePiece(GameState state, GameStateBuilder builder) {
  builder.queue
    ..remove(0)
    ..add(new Tetromino.Random());

  if (!builder.queue[0].points
      .every((point) => builder.board[point.x][point.y] == null))
    builder.gameOver = true;
}
